@@ -12,7 +12,7 @@ use AutoLoader qw/AUTOLOAD/ ;
 
 @ISA = qw(Tk::Derived Tk::Canvas);
 
-$VERSION = sprintf "%d.%03d", q$Revision: 1.18 $ =~ /(\d+)\.(\d+)/;
+$VERSION = sprintf "%d.%03d", q$Revision: 1.24 $ =~ /(\d+)\.(\d+)/;
 
 Tk::Widget->Construct('TreeGraph');
 
@@ -24,18 +24,20 @@ sub InitObject
     my $defc = $dw->parent->cget('-foreground');
     $dw->ConfigSpecs
       (
-       -shortcutColor => ['PASSIVE', undef, undef, 'orange'],
-       -shortcutStyle => ['PASSIVE', undef, undef, 'straight'],
-       -nodeColor     => ['PASSIVE', undef, undef, $defc],
-       -arrowColor    => ['PASSIVE', undef, undef, $defc],
-       -nodeTextColor => ['PASSIVE', undef, undef, $defc],
-       -labelColor    => ['PASSIVE', undef, undef, $defc],
-       -nodeTag       => ['PASSIVE', undef, undef, 1 ], # add node Id to text
+       -shortcutColor  => ['PASSIVE', undef, undef, 'orange'],
+       -shortcutStyle  => ['PASSIVE', undef, undef, 'straight'],
+       -animation      => ['PASSIVE', undef, undef, 0 ],
+       -nodeColor      => ['PASSIVE', undef, undef, $defc],
+       -nodeFill => ['PASSIVE', undef, undef, undef ],
+       -arrowColor     => ['PASSIVE', undef, undef, $defc],
+       -nodeTextColor  => ['PASSIVE', undef, undef, $defc],
+       -labelColor     => ['PASSIVE', undef, undef, $defc],
+       -nodeTag        => ['PASSIVE', undef, undef, 1 ], # add node Id to text
        # use this to tune the shape of nodes and arrows
-       -arrowDeltaY   => ['PASSIVE', undef, undef, 40 ],
+       -arrowDeltaY    => ['PASSIVE', undef, undef, 40 ],
        -branchSeparation => ['PASSIVE', undef, undef, 120 ],
-       -x_start       => ['PASSIVE', undef, undef,  40 ],
-       -y_start       => ['PASSIVE', undef, undef, 100 ]
+       -x_start        => ['PASSIVE', undef, undef,  40 ],
+       -y_start        => ['PASSIVE', undef, undef, 100 ]
       );
 
     # bind button <1> on nodes to select a version
@@ -68,7 +70,8 @@ Tk::TreeGraph - Tk widget to draw a tree in a Canvas
 
  my $mw = MainWindow-> new ;
 
- my $tg = $mw -> Scrolled('TreeGraph') ->pack(-expand => 1, -fill => 'both');
+ my $tg = $mw -> Scrolled('TreeGraph') ->
+  pack(-expand => 1, -fill => 'both');
 
  $tg -> addLabel (text => 'some tree');
 
@@ -132,7 +135,8 @@ Tk::TreeGraph - Tk widget to draw a tree in a Canvas
 
  # adjust scrolled area with some margin
  my @array = $tg->bbox("all") ;
- $tg->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);
+ $tg->configure(-scrollregion => 
+                [0, 0, $array[2] + 50, $array[3] + 50 ]);
 
  MainLoop ; # Tk's
 
@@ -211,51 +215,59 @@ will be faster.
 
 =over 4
 
-=item *
+=item -nodeColor: 
 
--nodeColor: Color of the node rectangle.
+Color of the node rectangle.
 
-=item *
+=item -nodeTextColor:
 
--nodeTextColor: Color of the text within the nodes
+Color of the text within the nodes
 
-=item *
+=item -labelColor
 
--labelColor
+Color of the label.
 
-=item *
+=item -arrowColor
 
--arrowColor
+Color of the arrow.
 
-=item *
+=item -shortcutColor:
 
--shortcutColor: Color of the shortcut arrow (default 'orange')
+Color of the shortcut arrow (default 'orange')
 
-=item *
+=item -shortcutStyle:
 
--shortcutStyle: Style of the shortcut arrow. The arrow can be drawn as 
-a 'straight' arrow or a 'spline'. (default 'straight')
+Style of the shortcut arrow. The arrow can be drawn as a 'straight'
+arrow or a 'spline'. (default 'straight')
 
-=item *
+=item -nodeTag:
 
--nodeTag: boolean. By default the nodeId is added at the beginning of
-the node text.
+Boolean. By default the nodeId is added at the beginning of the node
+text.
 
-=item *
+=item -arrowDeltaY:
 
--arrowDeltaY: length of direct arrows (downward). default 40
+Length of direct arrows (downward). default 40
 
-=item *
+=item -branchSeparation: 
 
--branchSeparation: minimum width between 2 branches of the tree (default 120) 
+Minimum width between 2 branches of the tree (default 120) 
 
-=item *
+=item -x_start:
 
--x_start: x coordinate of the root of the tree. (default 100)
+x coordinate of the root of the tree. (default 100)
 
-=item *
+=item -y_start: 
 
--y_start: y coordinate of the root of the tree.(default 100)
+y coordinate of the root of the tree.(default 100)
+
+=item -animation: 
+
+Delay (in ms), if delay is positive, TreeGraph will update the canvas
+widget any time the addNode method is called and wait delay ms. This
+enable to make pretty animation when drawing complex trees. This
+feature will work only if a scrollregion is set before calling
+addNode.
 
 =back
 
@@ -270,6 +282,8 @@ to call youself the addSlantedArrow or addDirectArrow methods.
 
 =head2 addNode(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -278,7 +292,7 @@ nodeId: string to identify this node.
 
 =item *
 
-text: text array ref. This text will be written inside the rectangle
+text: text string or array ref. This text will be written inside the rectangle
 
 =item *
 
@@ -294,11 +308,81 @@ Will add a new node (made of a rectangle with the text inside).
 Note that this method will add the nodeId on top of the passed text
 ('text' parameter).
 
+=head2 modifyNode(...)
+
+Parameters are:
+
+=over 4
+
+=item *
+
+nodeId: string to identify the node to modify (mandatory).
+
+=item *
+
+text: text string or array ref. This text will be overwritten inside
+the rectangle. Note that modifyNode will refuse to change the text if
+the new text has more lines than the older text (because the text
+would not fit in the rectangle).
+
+=item *
+
+nodeColor: new color of the outline of the rectangle.
+
+=item *
+
+nodeTextColor: new color of the text of the node.
+
+=item *
+
+nodeFill: new color filling the rectangle.
+
+=back
+
+Will modify an existing node. Note that the geometry of the node will not
+be changed.
+
+=head2 viewNode(nodeId)
+
+Will move the canvas so that the node is visible within the scrolled area.
+(do nothing if the scroll region is not defined TBD XXX)
+
+=head2 flashNode(...)
+
+Parameters are:
+
+=over 4
+
+=item *
+
+nodeId: string to identify the node to modify (mandatory).
+
+=item *
+
+time: time in ms between each flash (default 500 ms)
+
+=item *
+
+nodeColor: new color of the outline of the rectangle.
+
+=item *
+
+nodeTextColor: new color of the text of the node.
+
+=item *
+
+nodeFill: new color filling the rectangle.
+
+=back
+
+Will make an existing node flash. Calling a second time this method on
+a node will make the flashing stop.
+
 =head2 addDirectArrow(...)
 
 You can use this method if you want to change the default aspect of
 the direct arrow. In this case do not use the 'after' parameter of the
-addNode() method.
+addNode() method. Parameters are:
 
 =over 4
 
@@ -342,7 +426,7 @@ Add a new branch connecting node 'id' to node 'id2'.  Note that the
 
 =head2 addLabel(...)
 
-Put some text on the top of the graph.
+Put some text on the top of the graph. Parameters are:
 
 =over 4
 
@@ -354,6 +438,8 @@ text: text to be inserted on the top of the graph.
 
 =head2 addShortcutInfo(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -385,6 +471,8 @@ Clear the graph.
 
 =head2 nodeBind(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -409,6 +497,8 @@ with these parameters:
 
 =head2 arrowBind(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -502,6 +592,8 @@ inherit this widget.
 
 =head2 setArrow(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -518,6 +610,8 @@ the arrow is attached to.
 
 =head2 setNode()
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -539,6 +633,8 @@ if this function was used in a bind)
 
 =head2 toggleNode(...)
 
+Parameters are:
+
 =over 4
 
 =item *
@@ -554,22 +650,17 @@ pointer)
 
 Will toggle the node rectangle between 'color' and default.
 
-=head1 CAVEATS
-
-If you change the default size of the text written in the boxes, you
-will get mismatch between the size of the text and the size of the boxes.
-
 =head1 AUTHOR
 
 Dominique Dumont, Dominique_Dumont@grenoble.hp.com
 
-Copyright (c) 1998-2000 Dominique Dumont. All rights reserved.
+Copyright (c) 1998-2001 Dominique Dumont. All rights reserved.
 This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
 
 =head1 SEE ALSO
 
-perl(1), Tk(3), Tk::Canvas(3)
+L<Tk>, L<Tk::Canvas>
 
 =cut
 
@@ -601,7 +692,7 @@ sub clear
     my $dw = shift ;
     
     foreach (qw/arrow node nodeId tset xset shortcutFrom x y 
-             tree_start after limit/)
+             tree_start after limit column next_limit/)
       {
         delete $dw->{$_};
       }
@@ -643,7 +734,7 @@ sub checkOverlay
         $c++ ;
       }
 
-    #print "must move from column $old_c to $c)\n";
+    #print "must move from column $old_c to $c\n";
 
     my $b=$dw->{currentBranch} ;
     my $dx = $dw->cget('-branchSeparation')*($c-$old_c) ;
@@ -880,7 +971,8 @@ sub addNode
     my $dw = shift ;
     my %args = @_ ;
     my $nodeId = $args{nodeId} ;
-    my $textArrayRef = $args{text} ;
+    my $text = ref $args{text} ? join("\n",@{$args{text}}) : $args{text} ;
+    chomp $text;
 
     #print "Drawing node $nodeId\n";
     my $after = $args{after};
@@ -901,47 +993,51 @@ sub addNode
           }
       }
     
-    # initialiaztion
+    # initialization
     $dw->{tree_start}=$dw->cget('-x_start') unless defined $dw->{tree_start};
     
     # compute text to draw
-    my $text = $dw->cget('-nodeTag') ? 
-      join ("\n", $nodeId, @$textArrayRef)."\n" :
-        join ("\n", @$textArrayRef)."\n"  ;
+    $text = "$nodeId\n". $text if $dw->cget('-nodeTag') ;
 
     # first compute y coord 
     my $oldy = $dw->{y} || $dw->cget('-y_start');
     my $y = $oldy + 5 ; # give some breathing space 
 
-    # compute y according to the text drawn
-    $y += 14 if $dw->cget('-nodeTag') ; # add room for the node tag
-    $y += 14 * scalar(@$textArrayRef) + 10 ;
-
-    # check if we're not drawing over something
-    $dw->checkOverlay($y);
-
-    # then I can compute the x coordinate
+    # then compute the x coordinate
     my $branch_dx= $dw->cget('-branchSeparation');
     my $c = $dw->{column}{$dw->{currentBranch}} ;
     my $x = $branch_dx * $c + $dw->{tree_start}  ;
 
     # compute y coord
     # draw node text
-    my $defc = $dw->cget('-nodeTextColor');
+    my $defc = $args{nodeTextColor} || $args{-nodeTextColor} 
+      || $dw->cget('-nodeTextColor');
 
     my $tid = $dw->create('text', $x + $branch_dx/2 - 10, $oldy + 5, 
-                          -text=>$text,  -fill => $defc,
+                          -text => $text,  -fill => $defc,
                           qw/-justify center -anchor n -width 12c/,
                           -tags => ['node', $dw->{currentBranch}]) ;
 
+    # compute y according to the text drawn
+    my @box = $dw->bbox($tid) ;
+    $y = $box[3] + 5 ;
+
     # draw node rectangle
-    $defc = $dw->cget('-nodeColor');
+    $defc = $args{nodeColor} || $args{-nodeColor} || $dw->cget('-nodeColor');
+    my $bgc = $args{nodeFill} || $args{-nodeFill} 
+      || $dw->cget('-nodeFill');
+    my $newx = $x + $branch_dx   - 20 ;
     my $rid = $dw->create('rectangle',
                           $x , $oldy,
-                          $x + $branch_dx   - 20 , $y,
-                          -outline => $defc, -width => 2 , 
+                          $newx , $y,
+                          -outline => $defc, -width => 2 , -fill => $bgc ,
                           -tags => ['node', $dw->{currentBranch}]
                         ) ;
+    
+    $dw->raise($tid,$rid); # the text is hidden below if forgotten
+
+    # check if we have not drawn over something
+    $dw->checkOverlay($y);
 
     $dw -> {nodeId}{$tid}=$nodeId ; 
     $dw -> {nodeId}{$rid}=$nodeId ; # also stored
@@ -951,6 +1047,175 @@ sub addNode
 
     $dw->{x} = $x;
     $dw->{y} = $y ;
+
+    my $delay = $dw->cget('-animation') ;
+
+    if ($delay)
+      {
+        $dw->viewNode($nodeId) and $dw->after($delay);
+      }
+  }
+
+sub viewNode
+  {
+    my $dw = shift ;
+    my $nodeId = shift;
+
+    my @sc = $dw->cget('-scrollregion');
+
+    # even if the scrollregion is not defined, the array has 4 elements
+    # so i must tesst one of the elements of the array
+    return 0 unless (defined $sc[0]) ; # 
+
+    my @bb = $dw->bbox("all") ;
+
+    # adjust scroll region is something is outside of it
+    foreach my $i (0,1)
+      {
+        $sc[$i] = $bb[$i] - 50 if ($bb[$i] < $sc[$i]);
+      }
+    foreach my $i (2,3)
+      {
+        $sc[$i] = $bb[$i] + 50 if ($bb[$i] > $sc[$i]);
+      }
+
+    $dw->configure(-scrollregion => \@sc );
+
+    # get the rectangle coordinates
+    my $rectangle_id = $dw -> {node}{rectangle}{$nodeId};
+    my ($rx1,$ry1,$rx2,$ry2)= $dw->coords($rectangle_id);
+
+    # get the full canvas coordinates
+    my ($cx1,$cy1,$cx2,$cy2) = @sc;
+    # compute where is the node within this widget (fraction)
+    my ($xf1,$yf1,$xf2,$yf2) = 
+      ($rx1/($cx2-$cx1), $ry1/($cy2-$cy1), $rx2/($cx2-$cx1), $ry2/($cy2-$cy1));
+
+    # check if xf is out of view
+    my ($xvleft,$xvright) = $dw->xview ;
+    my $deltax = $xvright - $xvleft ;
+
+    # put the widget at the center (x wise)
+    $dw->xview(moveto=> (($xf1 + $xf2 - $deltax) / 2) )
+      if ($xf1< $xvleft or $xf2> $xvright);
+
+    # check if yf is out of view
+    my ($yvtop,$yvbottom) = $dw->yview ;
+    my $deltay = $yvbottom - $yvtop ;
+    $dw->yview(moveto=>(($yf1 + $yf2 - $deltay) / 2 ) )
+      if ($yf1 < $yvtop or $yf2 > $yvbottom);
+
+    # print "delay $x $y, fx $xf fy $yf\n" ;
+    $dw->idletasks;
+
+    return 1 ;
+  }
+
+sub modifyNode
+  {
+    my $dw = shift ;
+    my %args = @_ ;
+
+    my $nodeId = $args{nodeId} || $dw->getCurrentNodeId; # optional
+    croak "modifyNode: missing nodeId parameter" 
+      unless defined $args{nodeId};
+
+    my $rid = $dw->{node}{rectangle}{$nodeId} ; # retrieve id of rectangle
+    my $tid = $dw->{node}{text}{$nodeId} ; 
+
+    croak "modifyNode: unknown nodeId : $nodeId" unless defined $rid ;
+
+    if (defined $args{nodeColor})
+      {
+        $dw->itemconfigure($rid, -outline => $args{nodeColor}) ;
+      }
+    
+    if (defined $args{text})
+      {
+        my $text = ref $args{text} ? join("\n",$args{text}) : $args{text};
+        chomp $text ;
+
+        $text = $nodeId."\n$text" if $dw->cget('-nodeTag') ;
+        my $count = $text ;
+        $count =~ s/[^\n]// ;
+        
+        my $oldText = $dw->itemcget($tid, '-text' ) ;
+        $oldText =~ s/[^\n]// ;
+        
+        if (length($count) > length($oldText))
+          {
+            croak "modifyNode error: New text is longer than the old one. It will no fit in the node" ;
+          }
+        
+        $dw->itemconfigure($tid, -text => $text) ;
+      }
+    
+    if (defined $args{nodeTextColor})
+      {
+        $dw->itemconfigure($tid, -fill => $args{nodeTextColor}) ;
+      }
+    
+    if (defined $args{nodeFill})
+      {
+        $dw->itemconfigure($rid, -fill => $args{nodeFill}) ;
+      }
+    
+  }
+
+sub flashNode
+  {
+    my $dw = shift ;
+    my %args = @_ ;
+
+    my $nodeId = $args{nodeId} || $dw->getCurrentNodeId; # optional
+    croak "modifyNode: missing nodeId parameter" 
+      unless defined $args{nodeId};
+
+    my $rid = $dw->{node}{rectangle}{$nodeId} ; # retrieve id of rectangle
+    my $tid = $dw->{node}{text}{$nodeId} ; 
+
+    croak "modifyNode: unknown nodeId : $nodeId" unless defined $rid ;
+
+    if (defined $dw -> {node}{flash} and 
+        defined $dw -> {node}{flash}{$nodeId} and 
+        $dw -> {node}{flash}{$nodeId})
+      {
+        $dw -> {node}{flash}{$nodeId} = 0;
+        return ;
+      }
+    
+    $dw->viewNode($nodeId) ;
+
+    my $time = $args{time} || 500 ;
+    $dw -> {node}{flash}{$nodeId} = 1 ;
+
+    my $oldNodeColor = $dw->itemcget($rid, '-outline') ;
+    my $oldNodeTextColor = $dw->itemcget($tid, '-fill') ;
+    my $oldNodeFill = $dw->itemcget($rid, '-fill') ;
+
+    my $newNodeColor = $args{nodeColor} || $oldNodeColor ;
+    my $newNodeTextColor = $args{nodeTextColor} || $oldNodeTextColor ;
+    my $newNodeFill = $args{nodeFill} || $oldNodeFill ;
+
+    my ($on,$off) ;
+
+    $on = sub 
+      {
+        $dw->itemconfigure($rid, -outline => $newNodeColor) ;
+        $dw->itemconfigure($tid, -fill =>    $newNodeTextColor) ;
+        $dw->itemconfigure($rid, -fill =>    $newNodeFill) ;
+        $dw->after($time,$off) ; # always call off
+      };
+
+    $off = sub 
+      {
+        $dw->itemconfigure($rid, -outline => $oldNodeColor) ;
+        $dw->itemconfigure($tid, -fill =>    $oldNodeTextColor) ;
+        $dw->itemconfigure($rid, -fill =>    $oldNodeFill) ;
+        $dw->after($time,$on) if $dw -> {node}{flash}{$nodeId} ;
+      };
+
+    &$on ;
   }
 
 # will return with node Id
